# -*- coding: utf-8 -*-
#
# This file is part of simple-koji-ci project.
# Copyright (C) 2017  Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
# USA.
"""This provides classes and functions to work with the Fedora build system"""
from __future__ import absolute_import, print_function

from warnings import warn
import logging
import os
import random
import shutil
import string
import subprocess as sp
import tempfile
import threading
import time

import koji


_log = logging.getLogger(__name__)

_koji_session_lock = threading.RLock()


class Koji(object):

    def __init__(self, consumer, config):
        self.consumer = consumer
        self.config = config
        self.server = config['server']
        self.weburl = config['weburl']

        self.krb_principal = config['krb_principal']
        self.krb_keytab = config['krb_keytab']
        self.krb_ccache = config['krb_ccache']
        self.krb_sessionopts = config['krb_sessionopts']
        self.krb_proxyuser = config['krb_proxyuser']

        self.git_url = config['git_url']
        self.opts = config['opts']
        self.priority = config['priority']
        self.target_tags = config['target_tags']

    def session_maker(self):
        _log.info('Creating a new Koji session to %s', self.server)
        with _koji_session_lock:
            koji_session = koji.ClientSession(
                self.server, self.krb_sessionopts)
            result = koji_session.krb_login(
                principal=self.krb_principal,
                keytab=self.krb_keytab,
                ccache=self.krb_ccache,
                proxyuser=self.krb_proxyuser
            )
            if not result:
                _log.error('Koji kerberos authentication failed')
            return koji_session

    def url_for(self, task_id):
        return self.weburl + '/taskinfo?taskID=%i' % task_id

    @staticmethod
    def _unique_path(prefix):
        """ Create a unique path fragment.

        This is a copy and paste from /usr/bin/koji.
        """
        suffix = ''.join([
            random.choice(string.ascii_letters) for i in range(8)
        ])
        return '%s/%r.%s' % (prefix, time.time(), suffix)

    def upload_srpm(self, session, source):
        _log.info('Uploading {source} to koji'.format(source=source))
        serverdir = self._unique_path('cli-build')
        session.uploadWrapper(source, serverdir)
        return "%s/%s" % (serverdir, os.path.basename(source))

    def scratch_build(self, session, name, source, branch):
        remote = self.upload_srpm(session, source)
        # Make double sure we're only doing scrach builds
        opts = self.opts
        opts['scratch'] = True
        target = self.target_tags.get(branch, 'rawhide')
        _log.info('Intiating koji build for %r' % dict(
            name=name, target=target, source=remote, opts=opts))
        task_id = session.build(
            remote, target, opts, priority=self.priority)
        _log.info('Scratch build created for {name}: {url}'.format(
            name=name, url=self.url_for(task_id)))
        return task_id

    def handle(self, package, prid, branch_to):
        """ Main API entry point.

        Bumps the version on a package and requests a scratch build.

        Args:
            package (str): The package name.
            prid (int): The identifier of the pull-request to consider.

        Returns:
            int: koji task ID.
        """

        # Clone the package to a tempdir
        tmp = tempfile.mkdtemp(prefix='skc-', dir='/var/tmp')
        try:
            url = self.git_url.format(package=package)
            _log.info("Cloning %r to %r" % (url, tmp))
            sp.check_output(['git', 'clone', url, tmp], stderr=sp.STDOUT)
            sp.check_output(
                ['fedpkg', 'sources'], cwd=tmp, stderr=sp.STDOUT)
            sp.check_output([
                'git', 'fetch', 'origin', 'pull/%s/head:pr_%s' % (prid, prid)
                ], cwd=tmp, stderr=sp.STDOUT)
            sp.check_output(
                ['git', 'checkout', 'pr_%s' % prid],
                cwd=tmp, stderr=sp.STDOUT)

            specfile = tmp + '/' + package + '.spec'

            # Create the SRPM
            cmd = [
                'mock', '-r', '/etc/mock/simple_koji_ci.cfg',
                '--buildsrpm',
                '--define', '%_disable_source_fetch 0',
                '--spec', specfile,
                '--sources', tmp,
                '--resultdir', tmp,
            ]
            _log.info("Calling %r" % ' '.join(cmd))
            output = sp.check_output(
                cmd,
                cwd=tmp,
                stderr=sp.STDOUT,
            )
            _log.debug(output)

            # Retrieve the path to the srpm
            srpm = None
            for f in os.listdir(tmp):
                if f.endswith('.src.rpm'):
                    srpm = os.path.join(tmp, f)
                    break
            _log.debug("Got srpm %r" % srpm)

            # Kick off scratch build
            session = self.session_maker()
            _log.debug("Kicking koji build for: %r (%r)" % (package, srpm))
            task_id = self.scratch_build(session, package, srpm, branch_to)
            _log.debug("Task ID: %r" % task_id)

            return task_id
        except sp.CalledProcessError as err:
            _log.debug(err.cmd)
            _log.debug(err.returncode)
            _log.debug(err.output)
        finally:
            _log.debug("Removing %r" % tmp)
            shutil.rmtree(tmp, ignore_errors=True)
